module Benchmarks

package_path = dirname(normpath(joinpath(@__DIR__, "..")))
package_path ∉ LOAD_PATH && push!(LOAD_PATH, package_path)

using BenchmarkTools: BenchmarkGroup, @benchmarkable, judge, median, tune!
using Printf: @sprintf

using ValueConstraints:
    AbstractConstraint, All, Exactly, ExclusiveMaximum, ExclusiveMinimum, In, IsA, Not

# A helper comparator and corresponding constraint to benchmark the difference
# between vanilla and `ValueConstraints` implementations for non-trivial tasks.
#
# Mainly introduced to be able to test the efficacy of eager (compound)
# constraints that stop evaluating when doing so will not affect the outcome of
# the validation of the constraint, e.g. not evaluating the final two
# constraints if one out of five constraints needs to pass and the third
# constraint passes.
#
# The complexity of the `slowComparison` function is approximately the
# complexity where eager evaluation using `ValueConstraints` approaches the
# performance of short-circuiting vanilla logic
slowComparison(x::Number) = all(rand(1000) .> x)
struct SlowConstraint <: AbstractConstraint end
(constraint::SlowConstraint)(x::Number) = slowComparison(x)

function report(report_path=joinpath(@__DIR__, "README.md"))
    benchmarks = BenchmarkGroup(
        "ValueConstraints" => BenchmarkGroup(
            "Less Than" => @benchmarkable(isvalid(ExclusiveMaximum(rand()), rand())),
            "Constraint Inversion" =>
                @benchmarkable(isvalid(Not(ExclusiveMinimum(rand())), rand())),
            "Exhaustive Compound Multiple Type Constraint" => @benchmarkable(
                begin
                    test_value = rand() > 0.5 ? rand() : "Foo"
                    isvalid(All(In((1, "Foo")), IsA(String)), test_value)
                end
            ),
            "Exhaustive Less or Greater Than" => @benchmarkable(
                begin
                    test_value = rand()
                    isvalid(
                        Exactly{1}(ExclusiveMaximum(rand()), ExclusiveMinimum(rand())),
                        test_value,
                    )
                end
            ),
            "Exhaustive Slow Comparison or Less Than" => @benchmarkable(
                begin
                    test_value = rand()
                    isvalid(
                        Exactly{1}(SlowConstraint(), ExclusiveMaximum(rand())),
                        test_value,
                    )
                end
            ),
        ),
        "Vanilla" => BenchmarkGroup(
            "Less Than" => @benchmarkable(rand() < rand()),
            "Constraint Inversion" => @benchmarkable(!(rand() > rand())),
            "Exhaustive Compound Multiple Type Constraint" => @benchmarkable(
                begin
                    test_value = rand() > 0.5 ? rand() : "Foo"
                    in(test_value, (1, "Foo")) & isa(test_value, String)
                end
            ),
            "Exhaustive Less or Greater Than" => @benchmarkable(
                begin
                    test_value = rand()
                    (test_value < rand()) | (test_value > rand())
                end
            ),
            "Exhaustive Slow Comparison or Less Than" => @benchmarkable(
                begin
                    test_value = rand()
                    slowComparison(test_value) | (test_value < rand())
                end
            ),
        ),
    )

    tune!(benchmarks)
    results = run(benchmarks)
    judgement = judge(median.(values(results))...)

    open(report_path, "w") do report_stream
        write(
            report_stream,
            """
# Benchmarks

This report describes the performance of a set of constraints expressed
using the constructs in the `ValueConstraints` package versus the 'vanilla'
implementation, i.e. using basic Julia functions, using
[`BenchmarkTools`](https://juliaci.github.io/BenchmarkTools.jl/).

| Benchmark | Time | Memory |
| --------- | ---- | ------ |
""",
        )

        for benchmark in sort(collect(keys(judgement)))
            benchmark_judgement = judgement[benchmark]

            memory_judgement = @sprintf(
                "%s (%0.2f%%)",
                titlecase("$(benchmark_judgement.memory)"),
                benchmark_judgement.ratio.memory * 100
            )

            time_judgement = @sprintf(
                "%s (%0.2f%%)",
                titlecase("$(benchmark_judgement.time)"),
                benchmark_judgement.ratio.time * 100
            )

            write(
                report_stream,
                "| ",
                benchmark,
                " | ",
                time_judgement,
                " | ",
                memory_judgement,
                " |\n",
            )
        end

        write(
            report_stream,
            """

---

This report was generated by executing the following command from the root
of the project:

```sh
\$ julia --eval='include("./benchmarks/benchmarks.jl")' \\
        --eval='Benchmarks.report()' \\
        --project=benchmarks
```
""",
        )
    end

    return nothing
end

end
